/*
 *   jack-ladspa-host
 *    
 *   Copyright (C) Robert Ham 2002 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ladspa.h>
#include <dlfcn.h>

#include <glib.h>

#include "plugin.h"
#include "jack_rack.h"
#include "process.h"

#define CONTROL_FIFO_SIZE   256


/** connect up the ladspa instance's input buffers to the previous
    plugin's audio memory.  make sure to check that plugin->prev
    exists. */
void
plugin_connect_input_ports (plugin_t * plugin)
{
  if (!plugin || !plugin->prev)
    return;

  if (plugin->desc->mono)
    {
      plugin->descriptor->
	connect_port (plugin->holders[0]->instance,
		      plugin->desc->audio_port_indicies[IN_L],
		      plugin->prev->audio_memory[0]);
      plugin->descriptor->
	connect_port (plugin->holders[1]->instance,
		      plugin->desc->audio_port_indicies[IN_R],
		      plugin->prev->audio_memory[1]);
    }
  else
    {
      plugin->descriptor->
	connect_port (plugin->holders[0]->instance,
		      plugin->desc->audio_port_indicies[IN_L],
		      plugin->prev->audio_memory[0]);
      plugin->descriptor->
	connect_port (plugin->holders[0]->instance,
		      plugin->desc->audio_port_indicies[IN_R],
		      plugin->prev->audio_memory[1]);
    }
}

/** connect up a plugin's output ports to its own audio_memory output memory */
void
plugin_connect_output_ports (plugin_t * plugin)
{
  if (!plugin)
    return;
    
  if (plugin->desc->mono)
    {
      plugin->descriptor->
	connect_port (plugin->holders[0]->instance,
		      plugin->desc->audio_port_indicies[OUT_L], plugin->audio_memory[0]);
      plugin->descriptor->
	connect_port (plugin->holders[1]->instance,
		      plugin->desc->audio_port_indicies[OUT_R], plugin->audio_memory[1]);
    }
  else
    {
      plugin->descriptor->
	connect_port (plugin->holders[0]->instance,
		      plugin->desc->audio_port_indicies[OUT_L], plugin->audio_memory[0]);
      plugin->descriptor->
	connect_port (plugin->holders[0]->instance,
		      plugin->desc->audio_port_indicies[OUT_R], plugin->audio_memory[1]);
    }
}

void
process_add_plugin (process_info_t * procinfo, plugin_t * plugin)
{

  /* sort out list pointers */
  plugin->next = NULL;
  plugin->prev = procinfo->chain_end;

  if (procinfo->chain_end)
    procinfo->chain_end->next = plugin;
  else
    procinfo->chain = plugin;

  procinfo->chain_end = plugin;

}

/** find a plugin given its index in the chain */
static plugin_t *
process_get_plugin (process_info_t * procinfo, gint plugin_index)
{
  gint i = 0;
  plugin_t * plugin = procinfo->chain;

  while (i < plugin_index && plugin)
    {
      plugin = plugin->next;
      i++;
    }

  return plugin;
}

/** remove a plugin from the chain */
plugin_t *
process_remove_plugin (process_info_t * procinfo, gint plugin_index)
{
  plugin_t * plugin;

  /* find the plugin in the chain */
  plugin = process_get_plugin (procinfo, plugin_index);

  /* sort out chain pointers */
  if (plugin->prev)
    plugin->prev->next = plugin->next;
  else
    procinfo->chain = plugin->next;

  if (plugin->next)
    plugin->next->prev = plugin->prev;
  else
    procinfo->chain_end = plugin->prev;

  return plugin;
}

/** enable/disable a plugin */
void
process_ablise_plugin (process_info_t * procinfo, gint plugin_index, gint enable)
{
  plugin_t * plugin;

  plugin = process_get_plugin (procinfo, plugin_index);

  plugin->enabled = enable;
}

/** move a plugin up or down one place in the chain */
void
process_move_plugin (process_info_t * procinfo, gint plugin_index, gint up)
{

  /* specified plugin */
  plugin_t *plugin;
  /* other plugins in the chain */
  plugin_t *pp = NULL, *p, *n, *nn = NULL;

  plugin = process_get_plugin (procinfo, plugin_index);

  /* note that we should never recieve an illogical move request
     ie, there will always be at least 1 plugin before for an up
     request or 1 plugin after for a down request */

  /* these are pointers to the plugins surrounding the specified one:
     { pp, p, plugin, n, nn } which makes things much clearer than
     tptr, tptr2 etc */
  p = plugin->prev;
  if (p) pp = p->prev;
  n = plugin->next;
  if (n) nn = n->next;

  if (up)
    {
      if (!p)
	return;

      if (pp)
        pp->next = plugin;
      else
        procinfo->chain = plugin;

      p->next = n;
      p->prev = plugin;

      plugin->prev = pp;
      plugin->next = p;

      if (n)
	n->prev = p;
      else
	procinfo->chain_end = p;

    }
  else
    {
      if (!n)
	return;

      if (p)
	p->next = n;
      else
	procinfo->chain = n;

      n->prev = p;
      n->next = plugin;

      plugin->prev = n;
      plugin->next = nn;

      if (nn)
	nn->prev = plugin;
      else
	procinfo->chain_end = plugin;
    }
}

/** exchange an existing plugin for a newly created one */
plugin_t *
process_change_plugin (process_info_t * procinfo,
	               gint plugin_index, plugin_t * new_plugin)
{

  plugin_t *plugin;

  plugin = process_get_plugin (procinfo, plugin_index);

  new_plugin->next = plugin->next;
  new_plugin->prev = plugin->prev;

  if (plugin->prev)
    plugin->prev->next = new_plugin;
  else
    procinfo->chain = new_plugin;

  if (plugin->next)
    plugin->next->prev = new_plugin;
  else
    procinfo->chain_end = new_plugin;

  return plugin;
}


/******************************************
 ************* non RT stuff ***************
 ******************************************/


static int
plugin_init_instance (ladspa_holder_t * holder, plugin_t * plugin)
{
  unsigned long i;

  /* insantiate the plugin.. woo! */
  holder->instance = plugin->descriptor->
    instantiate (plugin->descriptor, sample_rate);

  if (!holder->instance)
    {
      fprintf (stderr,
	       "%s: could not initialise plugin '%s' from object file '%s'\n",
	       __FUNCTION__, plugin->desc->name, plugin->desc->object_file);
      return 1;
    }

  if (plugin->desc->control_port_count > 0)
    {
      holder->control_fifos =
	g_malloc (sizeof (lff_t) * plugin->desc->control_port_count);
      holder->control_memory =
	g_malloc (sizeof (LADSPA_Data) * plugin->desc->control_port_count);

      for (i = 0; i < plugin->desc->control_port_count; i++)
	{

	  /* init the fifo */
	  lff_init (holder->control_fifos + i, CONTROL_FIFO_SIZE,
		    sizeof (LADSPA_Data));

	  /* connect up the control memory */
/*      printf("%s: setting control port %ld of plugin '%s' to %p; plugin has %ld ports in file '%s'\n",
             __FUNCTION__, plugin->control_port_indicies[i],
             plugin->name, (void *) (holder->control_memory + i),
             plugin->descriptor->PortCount, plugin->object_file->filename);*/
	  plugin->descriptor->
	    connect_port (holder->instance, plugin->desc->control_port_indicies[i],
			  holder->control_memory + i);

	  /* set it to the default for now */
	  /* FIXME */
/*	  holder->control_memory[i] =
	    global_plugin_collection.default_plugin_settings[plugin->
							     collection_index]
	    [i];*/
	}
    }
  else
    {
      holder->control_fifos = NULL;
      holder->control_memory = NULL;
    }

  /* activate it */
  if (plugin->descriptor->activate)
    {
      plugin->descriptor->activate (holder->instance);
    }

  return 0;
}

static LADSPA_Data unused_control_port_output;

static int
plugin_open_object_file (plugin_t * plugin)
{
  const char * dlerr;
  LADSPA_Descriptor_Function get_descriptor;
  
  /* open the object file */
  plugin->dl_handle = dlopen (plugin->desc->object_file, RTLD_NOW|RTLD_GLOBAL);
  if (!plugin->dl_handle)
    {
      fprintf (stderr, "%s: error opening shared object file '%s': %s\n",
               __FUNCTION__, plugin->desc->object_file, dlerror());
      return 1;
    }
  
  
  /* get the get_descriptor function */
  dlerror (); /* clear the error report */
  
  get_descriptor = (LADSPA_Descriptor_Function)
    dlsym (plugin->dl_handle, "ladspa_descriptor");
  
  dlerr = dlerror();
  if (dlerr) {
    fprintf (stderr, "%s: error finding ladspa_descriptor symbol in object file '%s': %s\n",
             __FUNCTION__, plugin->desc->object_file, dlerr);
    dlclose (plugin->dl_handle);
    return 1;
  }
  
  plugin->descriptor = get_descriptor (plugin->desc->index);
  
  return 0;
}

plugin_t *
plugin_new (plugin_desc_t * desc)
{
  unsigned long i;
  int err;
  plugin_t * plugin;
  
  plugin = g_malloc (sizeof (plugin_t));

  plugin->desc = desc;
  plugin->enabled = FALSE;
  plugin->next = NULL;
  plugin->prev = NULL;

  plugin->audio_memory[0] = g_malloc (sizeof (LADSPA_Data) * buffer_size);
  plugin->audio_memory[1] = g_malloc (sizeof (LADSPA_Data) * buffer_size);
  
  err = plugin_open_object_file (plugin);
  if (err)
    {
      g_free (plugin->audio_memory[0]);
      g_free (plugin->audio_memory[1]);
      g_free (plugin);
      return NULL;
    }

  if (plugin->desc->mono)
    {

      plugin->holders[0] = g_malloc (sizeof (ladspa_holder_t));
      err = plugin_init_instance (plugin->holders[0], plugin);
      if (err)
	{
	  dlclose (plugin->dl_handle);
	  g_free (plugin->holders[0]);
	  g_free (plugin->audio_memory[0]);
	  g_free (plugin->audio_memory[1]);
	  g_free (plugin);
	  return NULL;
	}


      plugin->holders[1] = g_malloc (sizeof (ladspa_holder_t));
      err = plugin_init_instance (plugin->holders[1], plugin);
      if (err)
	{
	  dlclose (plugin->dl_handle);
	  if (plugin->holders[0]->control_fifos)
	    g_free (plugin->holders[0]->control_fifos);
	  if (plugin->holders[0]->control_memory)
	    g_free (plugin->holders[0]->control_fifos);
	  g_free (plugin->holders[1]);
	  g_free (plugin->holders[0]);
	  g_free (plugin->audio_memory[0]);
	  g_free (plugin->audio_memory[1]);
	  g_free (plugin);
	  return NULL;
	}

      /* connect up the audio memory */
      plugin->descriptor->connect_port (plugin->holders[0]->
					instance,
					plugin->desc->audio_port_indicies[OUT_L],
					plugin->audio_memory[0]);
      plugin->descriptor->connect_port (plugin->holders[1]->
					instance,
					plugin->desc->audio_port_indicies[OUT_R],
					plugin->audio_memory[1]);
    }
  else
    {
      plugin->holders[0] = g_malloc (sizeof (ladspa_holder_t));
      err = plugin_init_instance (plugin->holders[0], plugin);
      if (err)
	{
	  dlclose (plugin->dl_handle);
	  g_free (plugin->holders[0]);
	  g_free (plugin->audio_memory[0]);
	  g_free (plugin->audio_memory[1]);
	  g_free (plugin);
	  return NULL;
	}

      plugin->descriptor->connect_port (plugin->holders[0]->instance,
					plugin->desc->audio_port_indicies[OUT_L],
					plugin->audio_memory[0]);
      plugin->descriptor->connect_port (plugin->holders[0]->instance,
					plugin->desc->audio_port_indicies[OUT_R],
					plugin->audio_memory[1]);
    }

  /* connect up the control output ports (which we have to do) */
  for (i = 0; i < plugin->descriptor->PortCount; i++)
    {
      if (LADSPA_IS_PORT_CONTROL (plugin->descriptor->PortDescriptors[i]) &&
	  LADSPA_IS_PORT_OUTPUT (plugin->descriptor->PortDescriptors[i]))
	{
	  printf
	    ("%s: unused control output port at index %ld on plugin '%s'\n",
	     __FUNCTION__, i, plugin->desc->name);
	  if (plugin->desc->mono)
	    {
	      plugin->descriptor->connect_port (plugin->holders[0], i,
						&unused_control_port_output);
	      plugin->descriptor->connect_port (plugin->holders[1], i,
						&unused_control_port_output);
	    }
	  else
	    {
	      plugin->descriptor->connect_port (plugin->holders[0],
						i,
						&unused_control_port_output);
	    }
	}
    }

  return plugin;
}

void
plugin_destroy (plugin_t * plugin)
{
  int err;
/*  unsigned long i;*/

  /* deactivate and unload the plugins and free up their holders' memory */
  if (plugin->desc->mono)
    {
      if (plugin->descriptor->deactivate)
	{
	  plugin->descriptor->deactivate (plugin->holders[0]->instance);
	  plugin->descriptor->deactivate (plugin->holders[1]->instance);
	}
      plugin->descriptor->cleanup (plugin->holders[0]->
				       instance);
      plugin->descriptor->cleanup (plugin->holders[1]->
				       instance);

/*    FIXME: wtf is going on here?  these segfault
    for (i = 0; i < plugin->descriptor->PortCount; i++) {
       lff_free (plugin->holders[0]->control_fifos + i); 
       lff_free (plugin->holders[1]->control_fifos + i);
    }*/
      free (plugin->holders[0]->control_fifos);
      free (plugin->holders[1]->control_fifos);
      free (plugin->holders[0]->control_memory);
      free (plugin->holders[1]->control_memory);
      free (plugin->holders[0]);
      free (plugin->holders[1]);
    }
  else
    {
      if (plugin->descriptor->deactivate)
	{
	  plugin->descriptor->deactivate (plugin->holders[0]->
					      instance);
	}
      plugin->descriptor->cleanup (plugin->holders[0]->instance);

/*    for (i = 0; i < plugin->descriptor->PortCount; i++) {
      lff_free (plugin->holders[0]->control_fifos + i);
    }*/
      free (plugin->holders[0]->control_fifos);
      free (plugin->holders[0]->control_memory);
      free (plugin->holders[0]);
    }

  err = dlclose (plugin->dl_handle);
  if (err)
    {
      fprintf (stderr, "%s: error closing object file '%s' for plugin '%s': %s\n",
               __FUNCTION__, plugin->desc->object_file, plugin->desc->name, dlerror());
    }

  free (plugin->audio_memory[0]);
  free (plugin->audio_memory[1]);
  free (plugin);
}



/* EOF */


