/*
 *   jack-ladspa-host
 *    
 *   Copyright (C) Robert Ham 2002 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <jack/jack.h>
#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>

#include "process.h"
#include "control_message.h"
#include "lock_free_fifo.h"
#include "plugin.h"
#include "globals.h"

jack_nframes_t sample_rate;
jack_nframes_t buffer_size;

int process_control_messages (process_info_t * procinfo) {
  static int quitting = 0;
  ctrlmsg_t ctrlmsg;
  int err = 0;
  plugin_t * plugin;
  
  if (quitting) return quitting;
  
  while (lff_read (procinfo->gui_to_process, &ctrlmsg) == 0) {
  
    switch (ctrlmsg.type) {
    
      /* add a link to the end of the plugin chain */
      case CTRLMSG_ADD:
        plugin = ctrlmsg.pointer;
        process_add_plugin (procinfo, plugin);
        err = lff_write (procinfo->process_to_gui, &ctrlmsg);
        break;
        
      /* remove a link (plugin will be sent back) */
      case CTRLMSG_REMOVE:
        plugin = process_remove_plugin (procinfo, ctrlmsg.number);
        ctrlmsg.pointer = plugin;
        err = lff_write (procinfo->process_to_gui, &ctrlmsg);
        break;
        
      /* enable/disable a plugin */
      case CTRLMSG_ABLE:
        process_ablise_plugin (procinfo, ctrlmsg.number, GPOINTER_TO_INT(ctrlmsg.pointer));
        err = lff_write (procinfo->process_to_gui, &ctrlmsg);
        break;
      
      /* move a plugin up or down */  
      case CTRLMSG_MOVE:
/*        printf("%s: moving plugin %d %s\n", __FUNCTION__,
               ctrlmsg.number, (GPOINTER_TO_INT(ctrlmsg.pointer)) ? "up" : "down");*/
        process_move_plugin (procinfo, ctrlmsg.number,
                             GPOINTER_TO_INT(ctrlmsg.pointer));
        err = lff_write (procinfo->process_to_gui, &ctrlmsg);
        break;
              
      /* change an existing plugin for a new one (existing plugin
         will be sent back) */
      case CTRLMSG_CHANGE:
        plugin = ctrlmsg.pointer;
        plugin = process_change_plugin (procinfo, ctrlmsg.number, plugin);
        ctrlmsg.pointer = plugin;
        err = lff_write (procinfo->process_to_gui, &ctrlmsg);
        break;
        
      case CTRLMSG_QUIT:
        quitting = 1;
        return 1;
        break;
    }
    
    if (err) {
      fprintf (stderr, "%s: gui fifo is out of space\n", __FUNCTION__);
      return err;
    }
  }
  return 0;
}

/** process messages for plugins' control ports */
void process_control_port_messages (process_info_t * procinfo) {
  plugin_t * plugin;
  unsigned long control, copy;
  
  if (!procinfo->chain) return;
  
  for (plugin = procinfo->chain; plugin; plugin = plugin->next)
    if (plugin->desc->control_port_count > 0)
      for (control = 0; control < plugin->desc->control_port_count; control++)
        for (copy = 0; copy < plugin_copies (plugin->desc); copy++)
          while (lff_read (plugin->holders[copy]->control_fifos + control,
                           plugin->holders[copy]->control_memory + control) == 0);
}

int get_jack_buffers (process_info_t * procinfo, jack_nframes_t frames) {
  procinfo->jack_buffers[IN_L] = jack_port_get_buffer (procinfo->jack_ports[IN_L], frames);
  if (!procinfo->jack_buffers[IN_L]) {
    fprintf (stderr, "%s: no jack buffer for left input port\n", __FUNCTION__);
    return 1;
  }
  
  procinfo->jack_buffers[IN_R] = jack_port_get_buffer (procinfo->jack_ports[IN_R], frames);
  if (!procinfo->jack_buffers[IN_R]) {
    fprintf (stderr, "%s: no jack buffer for right input port\n", __FUNCTION__);
    return 1;
  }
  
  procinfo->jack_buffers[OUT_L] = jack_port_get_buffer (procinfo->jack_ports[OUT_L], frames);
  if (!procinfo->jack_buffers[OUT_L]) {
    fprintf (stderr, "%s: no jack buffer for left output port\n", __FUNCTION__);
    return 1;
  }
  
  procinfo->jack_buffers[OUT_R] = jack_port_get_buffer (procinfo->jack_ports[OUT_R], frames);
  if (!procinfo->jack_buffers[OUT_R]) {
    fprintf (stderr, "%s: no jack buffer for right output port\n", __FUNCTION__);
    return 1;
  }
  
  return 0;
}

plugin_t * get_first_enabled_plugin (process_info_t * procinfo) {
  plugin_t * first_enabled;
  
  if (!procinfo->chain) return NULL;

  for (first_enabled = procinfo->chain;
       first_enabled;
       first_enabled = first_enabled->next) {
    if (first_enabled->enabled) return first_enabled;
  }
 
  return NULL;
}

plugin_t * get_last_enabled_plugin (process_info_t * procinfo) {
  plugin_t * last_enabled;
  
  if (!procinfo->chain) return NULL;

  for (last_enabled = procinfo->chain_end;
       last_enabled;
       last_enabled = last_enabled->prev) {
    if (last_enabled->enabled) return last_enabled;
  }
  
  return NULL;
}

void connect_chain (process_info_t * procinfo, jack_nframes_t frames) {
  plugin_t * first_enabled, * last_enabled, * plugin;
  
  if (!procinfo->chain) return;
  
  first_enabled = get_first_enabled_plugin (procinfo);
  if (!first_enabled) return;
  
  last_enabled = get_last_enabled_plugin (procinfo);

  /* ensure that all the of the enabled plugins are connected to their memory */
  if (first_enabled != last_enabled) {
    plugin_connect_output_ports (first_enabled);
    plugin_connect_input_ports (last_enabled);
    for (plugin = first_enabled->next;
         plugin && (plugin != last_enabled);
         plugin = plugin->next) {
      plugin_connect_input_ports (plugin);
      plugin_connect_output_ports (plugin);
    }
  }

  /* input buffers for first plugin */
  if (first_enabled->desc->mono) {
    first_enabled->descriptor->
      connect_port (first_enabled->holders[0]->instance,
                    first_enabled->desc->audio_port_indicies[IN_L],
                    procinfo->jack_buffers[IN_L]);
    first_enabled->descriptor->
      connect_port (first_enabled->holders[1]->instance,
                    first_enabled->desc->audio_port_indicies[IN_R],
                    procinfo->jack_buffers[IN_R]);
  } else {
    first_enabled->descriptor->
      connect_port (first_enabled->holders[0]->instance,
                    first_enabled->desc->audio_port_indicies[IN_L],
                    procinfo->jack_buffers[IN_L]);
    first_enabled->descriptor->
      connect_port (first_enabled->holders[0]->instance,
                    first_enabled->desc->audio_port_indicies[IN_R],
                    procinfo->jack_buffers[IN_R]);
  }
  
  
  /* output buffers for last plugin */
  if (last_enabled->desc->mono) {
    last_enabled->descriptor->
      connect_port (last_enabled->holders[0]->instance,
                    last_enabled->desc->audio_port_indicies[OUT_L],
                    procinfo->jack_buffers[OUT_L]);
    last_enabled->descriptor->
      connect_port (last_enabled->holders[1]->instance,
                    last_enabled->desc->audio_port_indicies[OUT_R],
                    procinfo->jack_buffers[OUT_R]);
  } else {
    last_enabled->descriptor->
      connect_port (last_enabled->holders[0]->instance,
                    last_enabled->desc->audio_port_indicies[OUT_L],
                    procinfo->jack_buffers[OUT_L]);
    last_enabled->descriptor->
      connect_port (last_enabled->holders[0]->instance,
                    last_enabled->desc->audio_port_indicies[OUT_R],
                    procinfo->jack_buffers[OUT_R]);
  }
  

}

void process_chain (process_info_t * procinfo, jack_nframes_t frames) {
  plugin_t * first_enabled, * last_enabled = 0, * plugin;

  first_enabled = get_first_enabled_plugin (procinfo);
  
  /* no chain; just copy input to output */
  if (!procinfo->chain || !first_enabled) {
    memcpy (procinfo->jack_buffers[OUT_L], procinfo->jack_buffers[IN_L], sizeof(LADSPA_Data) * frames);
    memcpy (procinfo->jack_buffers[OUT_R], procinfo->jack_buffers[IN_R], sizeof(LADSPA_Data) * frames);
    return;
  }
  
  /* all past here is guaranteed to have at least 1 enabled plugin */

  last_enabled = get_last_enabled_plugin (procinfo);
  
  for (plugin = first_enabled;
       plugin;
       plugin = plugin->next) {
    if (plugin->enabled) {
    
      if (plugin->desc->mono) {
        plugin->descriptor->
          run (plugin->holders[0]->instance, frames);
        plugin->descriptor->
          run (plugin->holders[1]->instance, frames);
      } else {
        plugin->descriptor->
          run (plugin->holders[0]->instance, frames);
      }
      
      if (plugin == last_enabled) break;

    } else {
    
      /* copy the data through */
      memcpy (plugin->audio_memory[0],
              plugin->prev->audio_memory[0],
              sizeof(LADSPA_Data) * frames);
      memcpy (plugin->audio_memory[1],
              plugin->prev->audio_memory[1],
              sizeof(LADSPA_Data) * frames);
    }
  }
}

int process (jack_nframes_t frames, void * data) {
  int err, quitting;
  process_info_t * procinfo;
  
  procinfo = (process_info_t *) data;
  
  if (!procinfo) {
    fprintf (stderr, "%s: no process_info from jack!\n", __FUNCTION__);
    return 1;
  }
  
  quitting = process_control_messages (procinfo);
  
  if (quitting) return 1;
  
  process_control_port_messages (procinfo);
  
  err = get_jack_buffers (procinfo, frames);
  if (err) {
    fprintf(stderr, "%s: failed to get jack ports, not processing\n", __FUNCTION__);
    return 1;
  }
  
  connect_chain (procinfo, frames);
  
  process_chain (procinfo, frames);
  
  return 0;
}



/*******************************************
 ************** non RT stuff ***************
 *******************************************/

void
display_gtk_error (const char * message)
{
  GtkWidget * dialog;
  
  dialog = gtk_message_dialog_new (NULL,
                                   GTK_DIALOG_MODAL,
                                   GTK_MESSAGE_ERROR,
                                   GTK_BUTTONS_CLOSE,
                                   "Fatal error starting JACK Rack:\n\n%s",
                                   message);
  gtk_dialog_run (GTK_DIALOG (dialog));
}

process_info_t *
process_info_new (const char * client_name)
{
  process_info_t * procinfo;
  int i, err;
  int port = 0;
  const char ** jack_ports;
  char * port_name;
  int connect_inputs = 0;
  int connect_outputs = 0;

  struct
  {
    const char * name;
    int index;
    enum JackPortFlags flags;
  } port_map[]
  =
  {
    { "in_1", IN_R, JackPortIsInput },
    { "in_2", IN_L, JackPortIsInput },
    { "out_1", OUT_R, JackPortIsOutput },
    { "out_2", OUT_L, JackPortIsOutput }
  };
  
  
  procinfo = g_malloc (sizeof (process_info_t));
  
  procinfo->chain = NULL;
  procinfo->chain_end = NULL;
  
           
  procinfo->jack_client = jack_client_new (client_name);
  if (!procinfo->jack_client)
    {
      fprintf (stderr, "%s: could not create jack client; exiting\n", __FUNCTION__);
      fprintf (stderr, "%s:   is the jack server running?\n", __FUNCTION__);
      display_gtk_error ("could not create jack client; is the jackd server running?");
      exit (1);
    }
  
  /* register the ports */
  for (i = 0; i < 4; i++)
    {
      procinfo->jack_ports[port_map[i].index] =
        jack_port_register (procinfo->jack_client,
                            /* port name */    port_map[i].name,
                            /* port type */    JACK_DEFAULT_AUDIO_TYPE,
                            /* flags */        port_map[i].flags,
                            /* buffer size */  0);
    
      if (!procinfo->jack_ports[port_map[i].index])
        {
          fprintf (stderr, "%s: could not register port '%s'; exiting\n",
                   __FUNCTION__, port_map[i].name);
          display_gtk_error ("could not register jack port");
          exit (1);
        }
    }
  
  sample_rate = jack_get_sample_rate (procinfo->jack_client);
  buffer_size = jack_get_sample_rate (procinfo->jack_client);
  
  jack_set_process_callback (procinfo->jack_client, process, procinfo);
  
  /* connect up the ports */
#ifdef HAVE_LADCCA
  if (connect_inputs && !cca_enabled (global_cca_client))
#else
  if (connect_inputs)
#endif
    {
      jack_ports = jack_get_ports (procinfo->jack_client, NULL, NULL, JackPortIsOutput|JackPortIsPhysical);
      if (jack_ports)
        {
          for (i = 0; jack_ports[i] && port < 2; i++)
            {
              port_name = g_malloc (strlen (client_name) + 1 +
                                   strlen (port_map[port].name) + 1);
              sprintf (port_name, "%s:%s", client_name, port_map[port].name);
              
              err = jack_connect (procinfo->jack_client, jack_ports[i], port_name);
              
              if (err)
                fprintf (stderr, "%s: error connecting ports '%s' and '%s'\n", __FUNCTION__, port_name, jack_ports[i]);
              else
                port++;
              
            }
        }
    }
  
  port = 2;
#ifdef HAVE_LADCCA
  if (connect_outputs && !cca_enabled (global_cca_client))
#else
  if (connect_outputs)
#endif
    {
      jack_ports = jack_get_ports (procinfo->jack_client, NULL, NULL, JackPortIsInput|JackPortIsPhysical);
      if (jack_ports)
        {
          for (i = 0; jack_ports[i] && port < 4; i++)
            {
              port_name = g_malloc (strlen (client_name) + 1 +
                                   strlen (port_map[port].name) + 1);
              sprintf (port_name, "%s:%s", client_name, port_map[port].name);
              
              err = jack_connect (procinfo->jack_client, port_name, jack_ports[i]);
              
              if (err)
                fprintf (stderr, "%s: error connecting ports '%s' and '%s'\n", __FUNCTION__, port_name, jack_ports[i]);
              else
                port++;
              
            }
        }
    }
  
  /* init the fifos */
  procinfo->gui_to_process = g_malloc (sizeof (lff_t)); 
  procinfo->process_to_gui = g_malloc (sizeof (lff_t)); 
  lff_init (procinfo->gui_to_process, 32, sizeof (ctrlmsg_t));
  lff_init (procinfo->process_to_gui, 32, sizeof (ctrlmsg_t));
  

#ifdef HAVE_LADCCA
  /* sort out ladcca stuff */
  cca_jack_client_name (global_cca_client, client_name);
#endif


  return procinfo;
}

void
process_info_destroy (process_info_t * procinfo) {
  
  lff_free (procinfo->gui_to_process);
  lff_free (procinfo->process_to_gui);
  free (procinfo->gui_to_process);
  free (procinfo->process_to_gui);
}

